from .models import ChessBert, ChessGPT2, ChessT5
from ._loss import uci_loss, boardstate_loss, segmented_loss
from ._decode import segmented_argmax
from . import chars, uci, board

__all__ = [
  'ChessBert', 'ChessGPT2', 'ChessT5',
  'segmented_loss', 'uci_loss', 'boardstate_loss',
  'segmented_argmax',
  'chars', 'uci', 'board',
]


from jaxtyping import Float, Int
from haskellian import iter as I
import chess
import torch
from torch import Tensor, nn
from .._decode import segmented_argmax
from .._loss import segmented_loss

ENDS = [13*(i+1) for i in range(64)]

def loss(
  logits: Float[Tensor, 'batch seq_len 64*13'],
  labels: Int[Tensor, 'batch seq_len 64'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, '']:
  """Cross-entropy loss summed across the board squares"""
  losses = segmented_loss(logits, labels, loss=ce_loss, ends=ENDS)
  return sum(l for l in losses if not torch.isnan(l)) # type: ignore

def argmax_boards(logits: Float[torch.Tensor, 'B L 64*13']) -> list[list[chess.Board]]:
  ...
  # argmaxs = segmented_argmax(logits, ends=ENDS)
  # return I.ndmap(decode_board, argmaxs) # type: ignore



from typing import Sequence, TypedDict
import numpy as np
from haskellian import iter as I
import chess_utils as cu
from ._decode import decode_ucis

def accuracy(labs: Sequence[str], preds: Sequence[str]) -> float:
  """Proportion of correct predictions."""
  return sum(l == p for l, p in zip(labs, preds)) / len(labs)

def uci_accuracy(labs: Sequence[str], preds: Sequence[str]) -> float:
  """Proportion of correct UCI predictions."""
  return sum((l == p or l == p[:4]) for l, p in zip(labs, preds)) / len(labs)

class Metrics(TypedDict):
  uci_acc: float
  san_acc: float
  len_ratio: float

def metrics(uci_labs: Sequence[str], uci_preds: Sequence[str]) -> Metrics:
  san_labs = list(cu.ucis2sans(uci_labs))
  san_preds = list(decode_ucis(uci_preds))
  uci_acc = uci_accuracy(uci_labs, uci_preds)
  san_acc = accuracy(san_labs, san_preds)
  len_ratio = len(san_preds) / len(san_labs)
  return Metrics(uci_acc=uci_acc, san_acc=san_acc, len_ratio=len_ratio)

def batch_metrics(uci_labs: Sequence[Sequence[str]], uci_preds: Sequence[Sequence[str]]) -> Metrics:
  sample_metrics = [metrics(labs, preds) for labs, preds in zip(uci_labs, uci_preds)]
  return {
    k: np.mean(I.pluck(sample_metrics, k)) # type: ignore
    for k in sample_metrics[0].keys()
  } # type: ignore



from typing import TypedDict, NamedTuple, Sequence
from haskellian import iter as I
from jaxtyping import Int
import torch
from ...board import Sample as WordSample

class Sample(NamedTuple):
  input_ids: Int[torch.Tensor, 'L']
  labs: Int[torch.Tensor, 'L 5']

class Batch(TypedDict):
  input_ids: Int[torch.Tensor, 'B L']
  labs: Int[torch.Tensor, 'B L 5']
  attn_mask: Int[torch.Tensor, 'B L']

def sample_from(sample: WordSample) -> Sample:
  return Sample(torch.tensor(sample.input_ids), torch.tensor(sample.labs))

def collate_fn(batch: Sequence[Sample], pad_token_id: int, ignore_idx: int = -100) -> Batch:

  input_ids, labs = I.unzip(batch)
  batch_size = len(input_ids)
  maxlen = max(len(x) for x in input_ids)
  padded_input_ids = torch.full((batch_size, maxlen), fill_value=pad_token_id)
  padded_labs = torch.full((batch_size, maxlen, 64), fill_value=ignore_idx)
  attn_mask = torch.zeros((batch_size, maxlen))

  for i in range(batch_size):
    l = len(input_ids[i])
    padded_input_ids[i, :l] = input_ids[i]
    padded_labs[i, :l] = labs[i]
    attn_mask[i, :l] = 1

  return Batch(input_ids=padded_input_ids, labs=padded_labs, attn_mask=attn_mask)
